// Timer, V0.1
// Copyright 2011 by Frank Buss, fb@frank-buss.de

// include MCU specific things
#include <htc.h>

// for standard types like uint8_t
#include <stdint.h>

// chip configuration
__CONFIG(
	CP_OFF &  // Code protection off
	LVP_OFF &  // Low-Voltage Programming disabled
	BOREN_OFF &  // Brown-out Detect disabled
	MCLRE_OFF &  // RA5/MCLR pin function is digital input, MCLR internally tied to VDD
	PWRTE_ON &  // PWRT enabled
	WDTE_OFF &  // Watchdog Timer disabled
	FOSC_INTRC_CLKOUT  // low power oscillator 32 kHz
);

// define the oscillator frequency for the __delay_ms function
//#define _XTAL_FREQ 32768
#define _XTAL_FREQ 4000000

#define LCD_RS_BIT (1 << 2)  // RD2
#define LCD_E_BIT (1 << 3)  // RD3

uint8_t g_tick = 0;
uint8_t g_second = 0;
uint8_t g_minute = 0;
uint8_t g_secondDelay = 1;
uint8_t g_minuteDelay = 0;
uint8_t g_portA = 1 << 4;

uint8_t g_delaySelect = 0;

uint8_t g_init = 1;

void lcdSet(uint8_t control, uint8_t data)
{
	PORTD = (data & 0xf0) | control;
}

void lcdFunctionSetNibble(uint8_t data)
{
	lcdSet(0, data);
	__delay_ms(1);
	lcdSet(LCD_E_BIT, data);
	__delay_ms(1);
	lcdSet(0, 0);
	__delay_ms(1);
}

void lcdFunctionSet(uint8_t data)
{
	lcdFunctionSetNibble(data & 0xf0);
	lcdFunctionSetNibble((data << 4) & 0xf0);
}

void lcdRamSetNibble(uint8_t data)
{
	lcdSet(LCD_RS_BIT, data);
	__delay_ms(1);
	lcdSet(LCD_E_BIT | LCD_RS_BIT, data);
	__delay_ms(1);
	lcdSet(0, data);
	__delay_ms(1);
}

void lcdRamSet(uint8_t data)
{
	lcdRamSetNibble(data & 0xf0);
	lcdRamSetNibble((data << 4) & 0xf0);
}

void lcdInit()
{
	// init 4 bit mode
	lcdFunctionSetNibble(0b00110000);  // function set
	__delay_ms(50);
	lcdFunctionSetNibble(0b00110000);  // function set
	__delay_ms(50);
	lcdFunctionSetNibble(0b00110000);  // function set
	__delay_ms(50);
	lcdFunctionSetNibble(0b00100000);  // function set
	__delay_ms(50);

	// init display
	__delay_ms(50);
	lcdFunctionSet(0b00101000);  // function set
	__delay_ms(50);
	lcdFunctionSet(0b00101000);  // function set
	__delay_ms(50);
	lcdFunctionSet(0b00001100);  // display on/off
	__delay_ms(50);
	lcdFunctionSet(0b00000001);  // display clear
	__delay_ms(2);
	lcdFunctionSet(0b00000110);  // entry mode set
	__delay_ms(2);
}

void lcdClear()
{
	lcdFunctionSet(0b00000001);  // display clear
}

void lcdCursorOn()
{
	lcdFunctionSet(0b00001111);  // display on/off
	__delay_ms(50);
}

void setPosition(uint8_t x, uint8_t y)
{
	uint8_t position = x;
	if (y) position += 0x40;
	lcdFunctionSet(0b10000000 | position);  // set DDRAM address
}

void setChar(uint8_t x, uint8_t y, uint8_t c)
{
	setPosition(x, y);
	lcdRamSet(c);  // set character
}

void drawTextLine(uint8_t y, const char* text)
{
	uint8_t x = 0;
	while (*text) {
		setChar(x++, y, *text);
		text++;
	}
}

uint8_t isSelectPressed()
{
	return (PORTB & (1 << 7)) == 0;
}

uint8_t isSetPressed()
{
	return (PORTB & (1 << 6)) == 0;
}

// interrupt handler
static void interrupt
isr(void)
{
	static uint8_t selectPressed = 0;
	static uint8_t setPressed = 0;
	static uint8_t setRepeatCounter = 0;
	static uint8_t repeatCounterTestDelay = 0;
	static uint8_t triggerCounter = 0;
	uint8_t setAction = 0;
	uint8_t selectAction = 0;

	// clear interrupt flag
	TMR1IF = 0;

	PORTE = g_portA;
	g_portA ^= 0xff;
//	return;

	g_tick++;
	if (g_tick == 100) {
		g_tick = 0;
		g_second++;
		if (g_second == 60) {
			g_second = 0;
			g_minute++;
			if (g_minute == 60) {
				g_minute = 0;
			}
		}
		if (g_second == g_secondDelay && g_minute == g_minuteDelay) {
			g_second = 0;
			g_minute = 0;
			triggerCounter = 50;
			g_portA &= ~(1 << 4);
			PORTA = g_portA;
		}
	}

	if (triggerCounter > 0) {
		triggerCounter--;
		if (triggerCounter == 0) {
			g_portA |= 1 << 4;
			PORTA = g_portA;
		}
	}

	// read buttons and debounce
	if (isSelectPressed()) {
		if (!selectPressed) {
			selectPressed = 1;
			selectAction = 1;
		}
	} else {
		selectPressed = 0;
	}
	if (isSetPressed()) {
		if (repeatCounterTestDelay < 10) {
			repeatCounterTestDelay++;
		} else {
			repeatCounterTestDelay = 0;
			if (setRepeatCounter < 10) {
				setRepeatCounter++;
			} else {
				setAction = 1;
			}
		}
		if (!setPressed) {
			setPressed = 1;
			setAction = 1;
		}
	} else {
		setPressed = 0;
		setRepeatCounter = 0;
	}

	// select action
	if (selectAction) {
		g_delaySelect = 1 - g_delaySelect;
	}

	// set action
	if (setAction) {
		if (g_delaySelect) {
			g_minuteDelay++;
			if (g_minuteDelay == 100) {
				g_minuteDelay = 0;
			}
			g_second = 0;
			g_minute = 0;
		} else {
			g_secondDelay++;
			if (g_secondDelay == 60) {
				g_secondDelay = 0;
			}
			g_second = 0;
			g_minute = 0;
		}
	}
}

void drawNumber(uint8_t x, uint8_t y, uint8_t number)
{
	setChar(x, y, (number / 10) + '0');
	setChar(x + 1, y, (number % 10) + '0');
}

// main function
void main(void)
{
	// init ports
	PORTA = 1 << 4;
	PORTB = 0;

//	OSCCON = 1;

	// configure RB6 and RB7 to input, rest to output
	TRISA = 0;
	TRISB = (1 << 6) | (1 << 7);
	TRISD = 0;
	TRISE = 0;
/*	while (1) {
			PORTE = 255;
 			__delay_ms(5000);
			PORTE = 0;
 			__delay_ms(500);
	}*/

	// enable pull up on PORTB
	//OPTION_REG = 0;

	// Timer und Interrupt initialisieren
	// interrupt on TMR1
	PIE1 = 1;

	// period register
	PR2 = 255;

	// peripheral interrupt must be enabled for overflow interrupts
	PEIE = 1;  

	// start timer 1, prescale 16, postscale 4
//	T1CON = 0b00001011;
	T1CON = 0b00000001;

	// clear interrupt flag
//	TMR1IF = 0;

	GIE = 1;
//	OSCCON = 1;
//	while (1) {
//		asm("sleep");
//	}

	lcdInit();
	drawTextLine(0, "Impulsgeber V0.1");
	drawTextLine(1, "von Frank Buss");

	g_portA &= ~(1 << 4);
	PORTA = g_portA;
 	g_secondDelay = eeprom_read(0x00);
 	__delay_ms(20);
 	g_minuteDelay = eeprom_read(0x01);
 	__delay_ms(20);
	__delay_ms(5000);
	g_portA |= 1 << 4;
	PORTA = g_portA;

	lcdClear();
	drawTextLine(0, "Sekunden");
	drawTextLine(1, "Minuten");
	setPosition(0, 0);
	
	// global interrupt enable
	GIE = 1;
	g_init = 0;

	uint8_t oldSecond = 254;
	uint8_t oldMinute = 254;
	uint8_t oldDelaySelect = 254;
	uint8_t oldSecondDelay = 254;
	uint8_t oldMinuteDelay = 254;
	while (1) {
		if (oldSecondDelay != g_secondDelay) {
			oldSecondDelay = g_secondDelay;
			drawNumber(9, 0, oldSecondDelay);
		 	eeprom_write(0x00, oldSecondDelay);
 			__delay_ms(20);
		}
		if (oldMinuteDelay != g_minuteDelay) {
			oldMinuteDelay = g_minuteDelay;
			drawNumber(9, 1, oldMinuteDelay);
		 	eeprom_write(0x01, oldMinuteDelay);
 			__delay_ms(20);
		}
		if (oldSecond != g_second) {
			oldSecond = g_second;
			drawNumber(14, 0, oldSecond);
		}
		if (oldMinute != g_minute) {
			oldMinute = g_minute;
			drawNumber(14, 1, oldMinute);
		}
		if (oldDelaySelect != g_delaySelect) {
			oldDelaySelect = g_delaySelect;
			if (oldDelaySelect) {
				setChar(12, 0, ' ');
				setChar(12, 1, '*');
			} else {
				setChar(12, 0, '*');
				setChar(12, 1, ' ');
			}
		}
	}

	// count keyboard press events
	uint8_t counter = 0;
	uint8_t waitForHigh = 0;
	while (1) {
		// update LEDs with counter value
		PORTB = counter & 0x0f;

		// detect falling edge
		if (waitForHigh) {
			if (PORTBbits.RB4 == 1) {
				waitForHigh = 0;
			}
		} else {
			if (PORTBbits.RB4 == 0) {
				counter++;
				waitForHigh = 1;
			}
		}

		// delay to avoid bouncing
		__delay_ms(50);
	}
}
